/* Low level interface to ptrace, for the remote server for GDB.
   Copyright 1995, 1996, 1998, 1999, 2000, 2001, 2002, 2003, 2004
   Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD: src/gnu/usr.bin/gdb/gdbserver/fbsd-low.c,v 1.2.4.1 2007/12/04 18:00:17 obrien Exp $");

#include "server.h"
#include "fbsd-low.h"

#include <sys/wait.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/ptrace.h>
#include <sys/jnx/_timeval.h>
#include <sys/jnx/resource.h>
#include <sys/user.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/syscall.h>

struct pending_signals
{
  int signal;
  struct pending_signals *prev;
};

#define PTRACE_ARG3_TYPE caddr_t
#define PTRACE_XFER_TYPE int

int using_threads;
int debug_threads;

struct process_info*
get_thread_process (struct thread_info* thread)
{
  return (struct process_info*) inferior_target_data (thread);
}

static void
print_grs (CORE_ADDR* buf, long lwp)
{
  int i;
  fprintf (stderr, "%ld: ", lwp);
  for (i = 0; i < the_low_target.num_regs; i++)
    fprintf (stderr, "0x%lx ", buf[i]);
  fprintf (stderr, "\n");
}
static void
print_thread_grs (struct inferior_list_entry *entry)
{
  int res, i;
  CORE_ADDR* buf;
  struct thread_info *thread = (struct thread_info *) entry;
  struct regset_info *regset = target_regsets;

  while (regset->type != GENERAL_REGS && regset->size >=0)
    regset++;

  if (regset->type == GENERAL_REGS)
    {
      buf = malloc (regset->size);
      res = ptrace (regset->get_request,
		    GET_LWP (thread_to_gdb_id (thread)),
		    (PTRACE_ARG3_TYPE) buf, 0);
      if (res < 0)
	{
	  char s[256];
	  sprintf (s, "ptrace(regsets_fetch_inferior_registers) PID=%d",
		   GET_LWP (thread_to_gdb_id (thread)));
	  perror (s);
	}

      print_grs (buf, GET_LWP (thread_to_gdb_id (thread)));
      free (buf);
    }
}

static print_all_thread_grs (void)
{
  for_each_inferior (&all_threads, print_thread_grs);
}

/* This function should only be called if the process got a SIGTRAP.
   The SIGTRAP could mean several things.

   On i386, where decr_pc_after_break is non-zero:
   If we were single-stepping this process using PT_STEP,
   we will get only the one SIGTRAP (even if the instruction we
   stepped over was a breakpoint).  The value of $eip will be the
   next instruction.
   If we continue the process using PTRACE_CONT, we will get a
   SIGTRAP when we hit a breakpoint.  The value of $eip will be
   the instruction after the breakpoint (i.e. needs to be
   decremented).  If we report the SIGTRAP to GDB, we must also
   report the undecremented PC.  If we cancel the SIGTRAP, we
   must resume at the decremented PC.

   (Presumably, not yet tested) On a non-decr_pc_after_break machine
   with hardware or kernel single-step:
   If we single-step over a breakpoint instruction, our PC will
   point at the following instruction.  If we continue and hit a
   breakpoint instruction, our PC will point at the breakpoint
   instruction.  */
static CORE_ADDR
get_stop_pc (void)
{
  CORE_ADDR stop_pc = (*the_low_target.get_pc) ();
  struct process_info* process = get_thread_process (current_inferior);

  if (process->stepping)
    return stop_pc;
  else
    return stop_pc - the_low_target.decr_pc_after_break;
}

static void
fbsd_delete_thread (struct thread_info* thread)
{
  struct process_info* process = get_thread_process (thread);
  if (process)
    free (process);

  if (debug_threads)
    fprintf (stderr, "%ld: delete thread.\n", 
	     GET_LWP (thread_to_gdb_id (thread)));

  remove_thread (thread);
}

int
fbsd_create_inferior (char *program, char **args)
{
  struct process_info* process;
  int pid;

  pid = vfork ();
  if (pid < 0)
    {
      perror_with_name ("vfork");
    }

  if (pid == 0)
    {
      ptrace (PT_TRACE_ME, 0, 0, 0);
      setpgid (0, 0);
      execv (program, args);

      fprintf (stderr, "Cannot exec %s: %s.\n", program,
	       strerror (errno));
      fflush (stderr);
      _exit (0177);
    }

  process = (struct process_info*) malloc (sizeof(struct process_info));
  memset (process, 0, sizeof(struct process_info));
  add_thread (BUILD_LWP (pid, pid), process);
  return pid;
}

int
fbsd_attach (unsigned long pid)
{
  struct process_info* process;
  if (ptrace (PT_ATTACH, pid, 0, 0) != 0)
    {
      fprintf (stderr, "Cannot attach to process %d: %s (%d)\n", pid,
	       strerror (errno), errno);
      fflush (stderr);
      return 1;
    }

  process = (struct process_info*) malloc (sizeof(struct process_info));
  memset (process, 0, sizeof(struct process_info));
  add_thread (BUILD_LWP (pid, pid), process);
  return 0;
}

void
fbsd_kill (void)
{
  int ret, wstat;
  do
    {
      ptrace (PT_KILL, ptid_get_pid (all_threads.head->id), 0, 0);
      while(1)
	{
	  ret = waitpid (ptid_get_pid (all_threads.head->id), &wstat, WNOHANG);
	  if (ret == -1)
	    {
	      if (errno != ECHILD)
		{
		  perror_with_name ("waitpid");
		}
	    }
	  else if (ret > 0)
	    {
	      break;
	    }
	  usleep (1000);
	}
    } while (WIFSTOPPED (wstat));
}

void
fbsd_detach (void)
{
  ptrace (PT_DETACH, ptid_get_pid (all_threads.head->id), 0, 0);
}

int
fbsd_thread_alive (ptid_t ptid)
{
  if (find_inferior_id (&all_threads, ptid) != NULL)
    return 1;
  else
    return 0;
}

/* Return true if this process stopped at a breakpoint which
   no longer appears to be inserted.  Also adjust the PC
   appropriately to resume where the breakpoint used to be.  */
static int
check_removed_breakpoint (struct thread_info *thread)
{
  CORE_ADDR stop_pc;
  struct thread_info *saved_inferior;
  struct process_info* process = get_thread_process (thread);

  if (process->pending_is_breakpoint == 0)
    return 0;

  saved_inferior = current_inferior;
  current_inferior = thread;

  stop_pc = get_stop_pc ();

  /* If the PC has changed since we stopped, then we shouldn't do
     anything.  This happens if, for instance, GDB handled the
     decr_pc_after_break subtraction itself.  */
  if (stop_pc != process->pending_stop_pc)
    {
      if (debug_threads)
	fprintf (stderr, "%ld: ignoring, PC was changed.\n",
		 GET_LWP (thread_to_gdb_id (current_inferior)));

      process->pending_is_breakpoint = 0;
      current_inferior = saved_inferior;
      return 0;
    }

  /* If the breakpoint is still there, we will report hitting it.  */
  if ((*the_low_target.breakpoint_at) (stop_pc))
    {
      if (debug_threads)
	fprintf (stderr, "%ld: ignoring, breakpoint is still present.\n",
		 GET_LWP (thread_to_gdb_id (current_inferior)));

      current_inferior = saved_inferior;
      return 0;
    }

  if (debug_threads)
    fprintf (stderr, "%ld: removed breakpoint.\n",
	     GET_LWP (thread_to_gdb_id (current_inferior)));

  /* For decr_pc_after_break targets, here is where we perform the
     decrement.  We go immediately from this function to resuming,
     and can not safely call get_stop_pc () again.  */
  if (the_low_target.set_pc != NULL)
    (*the_low_target.set_pc) (stop_pc);

  /* We consumed the pending SIGTRAP.  */
  process->pending_is_breakpoint = 0;
  process->status_pending_p = 0;
  process->status_pending = 0;

  current_inferior = saved_inferior;
  return 1;
}

static void
fbsd_resume_one_thread (struct thread_info *thread, int step, 
		       int signal, int next_continue)
{
  struct thread_info *saved_inferior;
  struct process_info* process;
  int resume_all;

  /* If step is followed by continue, we need to resume all threads */
  if (thread && step)
    {
      resume_all = next_continue;
    }
  else
    {
      resume_all = 1;
      thread = thread ? thread : get_last_step_thread();
    }

  if (debug_threads)
    fprintf (stderr, "%ld: resuming, step = %d, signal = %d, "
	     "resume_all = %d.\n", 
	     GET_LWP (thread_to_gdb_id (thread)), step, signal, resume_all);

  process = get_thread_process (thread);
  if (process->stopped == 0)
    {
      if (debug_threads)
	fprintf (stderr, "%ld: no need to resume, already running.\n", 
		 GET_LWP (thread_to_gdb_id (thread)));
      return;
    }

  /* If we have pending signals or status, and a new signal, enqueue the
     signal.  Also enqueue the signal if we are waiting to reinsert a
     breakpoint; it will be picked up again below.  */
  if (signal != 0
      && (process->status_pending_p || process->pending_signals != NULL
	  || process->bp_reinsert != 0))
    {
      struct pending_signals *p_sig;
      p_sig = malloc (sizeof (*p_sig));
      p_sig->prev = process->pending_signals;
      p_sig->signal = signal;
      process->pending_signals = p_sig;

      if (debug_threads)
	fprintf (stderr, "%ld: added pending signal.\n", 
		 GET_LWP (thread_to_gdb_id (thread)));
    }

  if (process->status_pending_p && !check_removed_breakpoint (thread))
    {
      if (debug_threads)
	fprintf (stderr, "%ld: status pending, return.\n", 
		 GET_LWP (thread_to_gdb_id (thread)));
      return;
    }

  saved_inferior = current_inferior;
  current_inferior = thread;

  /* TODO: This bit needs some thinking about.  If we get a signal that
     we must report while a single-step reinsert is still pending,
     we often end up resuming the thread.  It might be better to
     (ew) allow a stack of pending events; then we could be sure that
     the reinsert happened right away and not lose any signals.

     Making this stack would also shrink the window in which breakpoints are
     uninserted (see comment in fbsd_wait_for_process) but not enough for
     complete correctness, so it won't solve that problem.  It may be
     worthwhile just to solve this one, however.  */
  if (process->bp_reinsert != 0)
    {
      if (debug_threads)
	fprintf (stderr, "%ld: bp reinsert.\n", 
		 GET_LWP (thread_to_gdb_id (thread)));

      if (step == 0)
	fprintf (stderr, "BAD - reinserting but not stepping.\n");
      step = 1;

      /* Postpone any pending signal.  It was enqueued above.  */
      signal = 0;
    }

  check_removed_breakpoint (thread);

  /* If we have pending signals, consume one unless we are trying to reinsert
     a breakpoint.  */
  if (process->pending_signals != NULL && process->bp_reinsert == 0)
    {
      struct pending_signals **p_sig;

      p_sig = &process->pending_signals;
      while ((*p_sig)->prev != NULL)
	p_sig = &(*p_sig)->prev;

      signal = (*p_sig)->signal;
      free (*p_sig);
      *p_sig = NULL;

      if (debug_threads)
	fprintf (stderr, "l%d: consumed pending signal.\n", 
		 GET_LWP (thread_to_gdb_id (thread)));
    }

  regcache_invalidate ();

  if (using_threads)
    fbsd_suspend_resume (resume_all, thread_to_gdb_id (thread));

  if (debug_threads)
    fprintf (stderr, "%ld: continue, step = %d, signal = %d.\n", 
	     GET_LWP (thread_to_gdb_id (thread)), step, signal);

  errno = 0;
  process->stopped = 0;
  process->stepping = step;
  ptrace (step ? PT_SETSTEP : PT_CLEARSTEP, 
	  GET_LWP (thread_to_gdb_id (thread)), 
	  (PTRACE_ARG3_TYPE) 1, signal);
  ptrace (PT_CONTINUE, GET_PID (thread_to_gdb_id (thread)), 
	  (PTRACE_ARG3_TYPE) 1, signal);

  current_inferior = saved_inferior;
  if (errno)
    perror_with_name ("ptrace");
}

void
fbsd_resume (struct thread_resume *resume_info)
{
  struct thread_info* thread;
  int next_continue = 0;
  int ndx = 0;

  do
    {
      if (resume_info[ndx].leave_stopped || next_continue)
	continue;

      thread = gdb_id_to_thread (resume_info[ndx].thread);
      next_continue = 
	(thread && !resume_info[ndx + 1].leave_stopped &&
	 gdb_id_to_thread (resume_info[ndx + 1].thread) == NULL) ? 
	1 : 0;

      fbsd_resume_one_thread (thread, resume_info[ndx].step, 
			      resume_info[ndx].sig, next_continue);
    }
  while (!ptid_equal (resume_info[ndx++].thread, minus_one_ptid));
}

static ptid_t
fbsd_wait_for_process (int *wstatp)
{
  int ret;
  struct ptrace_lwpinfo pl;
  ptid_t ptid = minus_one_ptid;
  int to_wait_for = -1;

  while (1)
    {
      ret = waitpid (to_wait_for, wstatp, WNOHANG);

      if (ret == -1)
	{
	  if (errno != ECHILD)
	    perror_with_name ("waitpid");
	}
      else if (ret > 0)
	break;

      usleep (1000);
    }

  if (debug_threads)
    fprintf (stderr, "%ld: WIFCONTINUED = %d, WIFEXITED = %d, "
	     "WIFSIGNALED = %d, WIFSTOPPED = %d\n", 
	     ret, WIFCONTINUED (*wstatp), WIFEXITED (*wstatp), 
	     WIFSIGNALED (*wstatp), WIFSTOPPED (*wstatp));

  if (ret > 0 && WIFSTOPPED (*wstatp))
    {
      if (ptrace (PT_LWPINFO, ret, (caddr_t) &pl, sizeof (pl)))
	perror_with_name ("PT_LWPINFO");
      ptid = BUILD_LWP ((long) pl.pl_lwpid, GET_PID (all_threads.head->id));
      fbsd_attach_lwp (ptid);

      if (debug_threads)
	fprintf (stderr, "%ld: event %d LWP = %ld.\n", 
		 ret, pl.pl_event, pl.pl_lwpid);
    }
  else
    {
      ptid = all_threads.head->id;
      if (debug_threads)
	fprintf (stderr, "%ld: pick main thread.\n", ret);
    }

  if (debug_threads)
    fprintf (stderr, "%ld: stopped with signal %d\n", 
	     GET_LWP (ptid), WSTOPSIG (*wstatp));

  return ptid;
}

static int
fbsd_wait_for_event (void)
{
  struct process_info* process;
  CORE_ADDR stop_pc;
  int wstat;

  while (1)
    {
      ptid_t ptid = fbsd_wait_for_process (&wstat);

      if (WIFEXITED(wstat))
	{
	  if (debug_threads)
	    fprintf (stderr, "%ld: exited.\n", GET_LWP (ptid));
	  return wstat;
	}

      current_inferior = (struct thread_info *)
	find_inferior_id (&all_threads, ptid);
      process = get_thread_process (current_inferior);

      process->stopped = 1;
      process->pending_is_breakpoint = 0;

      if (using_threads)
	{
	  if (!WIFSTOPPED (wstat))
	    {
	      /* If the last thread is exiting, just return.  */
	      if (all_threads.head == all_threads.tail)
		{
		  if (debug_threads)
		    fprintf (stderr, "Last thread exited.\n");
		  return wstat;
		}

	      if (debug_threads)
		fprintf (stderr, 
			 "%ld: delete thread and wait for more events.\n", 
			 GET_LWP (ptid));

	      dead_thread_notify (GET_LWP (thread_to_gdb_id (current_inferior)));
	      fbsd_delete_thread (current_inferior);
	      current_inferior = (struct thread_info *) all_threads.head;

	      /* Wait for a more interesting event.  */
	      continue;
	    }
	}

      /* FIXME drow/2002-06-09: Get signal numbers from the inferior's
	 thread library?  */
      if (WIFSTOPPED (wstat) &&
	  (WSTOPSIG (wstat) == 32 || WSTOPSIG (wstat) == 33))
	{
	  if (debug_threads)
	    fprintf (stderr, "%ld: ignored signal %d.\n",
		     GET_LWP (thread_to_gdb_id (current_inferior)),
		     WSTOPSIG (wstat));
	  
	  fbsd_resume_one_thread (current_inferior, process->stepping,
				  WSTOPSIG (wstat), 0);
	  continue;
	}

      /* If this event was not handled above, and is not a SIGTRAP, report
	 it.  */
      if (!WIFSTOPPED (wstat) || WSTOPSIG (wstat) != SIGTRAP)
	{
	  if (debug_threads)
	    fprintf (stderr, "%ld: !WIFSTOPPED or !SIGTRAP.\n", GET_LWP (ptid));
	  return wstat;
	}

      /* If this target does not support breakpoints, we simply report the
	 SIGTRAP; it's of no concern to us.  */
      if (the_low_target.get_pc == NULL)
	return wstat;

      stop_pc = get_stop_pc ();

      /* bp_reinsert will only be set if we were single-stepping.
	 Notice that we will resume the process after hitting
	 a gdbserver breakpoint; single-stepping to/over one
	 is not supported (yet).  */
      if (process->bp_reinsert != 0)
	{
	  if (debug_threads)
	    fprintf (stderr, "%ld: breakpoint re-insert.\n", 
		     GET_LWP (thread_to_gdb_id (current_inferior)));

	  reinsert_breakpoint (process->bp_reinsert);
	  process->bp_reinsert = 0;

	  /* Clear the single-stepping flag and SIGTRAP as we resume.  */
	  fbsd_resume_one_thread (current_inferior, 0, 0, 0);
	  continue;
	}

      if (check_breakpoints (stop_pc) != 0)
	{
	  if (debug_threads)
	    fprintf (stderr, "%ld: hit own breakpoint.\n", 
		     GET_LWP (thread_to_gdb_id (current_inferior)));

	  /* We hit one of our own breakpoints.  We mark it as a pending
	     breakpoint, so that check_removed_breakpoint () will do the PC
	     adjustment for us at the appropriate time.  */
	  process->pending_is_breakpoint = 1;
	  process->pending_stop_pc = stop_pc;

	  /* Now we need to put the breakpoint back.  We continue in the event
	     loop instead of simply replacing the breakpoint right away,
	     in order to not lose signals sent to the thread that hit the
	     breakpoint.  Unfortunately this increases the window where another
	     thread could sneak past the removed breakpoint.  For the current
	     use of server-side breakpoints (thread creation) this is
	     acceptable; but it needs to be considered before this breakpoint
	     mechanism can be used in more general ways.  For some breakpoints
	     it may be necessary to stop all other threads, but that should
	     be avoided where possible.

	     If breakpoint_reinsert_addr is NULL, that means that we can
	     use PT_STEP on this platform.  Uninsert the breakpoint,
	     mark it for reinsertion, and single-step.

	     Otherwise, call the target function to figure out where we need
	     our temporary breakpoint, create it, and continue executing this
	     process.  */
	  if (the_low_target.breakpoint_reinsert_addr == NULL)
	    {
	      process->bp_reinsert = stop_pc;
	      uninsert_breakpoint (stop_pc);
	      fbsd_resume_one_thread (current_inferior, 1, 0, 0);
	    }
	  else
	    {
	      reinsert_breakpoint_by_bp
		(stop_pc, (*the_low_target.breakpoint_reinsert_addr) ());
	      fbsd_resume_one_thread (current_inferior, 0, 0, 0);
	    }

	  continue;
	}

      /* If we were single-stepping, we definitely want to report the
	 SIGTRAP.  The single-step operation has completed, so also
         clear the stepping flag; in general this does not matter,
	 because the SIGTRAP will be reported to the client, which
	 will give us a new action for this thread, but clear it for
	 consistency anyway.  It's safe to clear the stepping flag
         because the only consumer of get_stop_pc () after this point
	 is check_removed_breakpoint, and pending_is_breakpoint is not
	 set.  It might be wiser to use a step_completed flag instead.  */
      if (process->stepping)
	{
	  if (debug_threads)
	    fprintf (stderr, "%ld: stepping, report SIGTRAP.\n", 
		     GET_LWP (thread_to_gdb_id (current_inferior)));

	  process->stepping = 0;
	  return wstat;
	}

      /* A SIGTRAP that we can't explain.  It may have been a breakpoint.
	 Check if it is a breakpoint, and if so mark the process information
	 accordingly.  This will handle both the necessary fiddling with the
	 PC on decr_pc_after_break targets and suppressing extra threads
	 hitting a breakpoint if two hit it at once and then GDB removes it
	 after the first is reported.  Arguably it would be better to report
	 multiple threads hitting breakpoints simultaneously, but the current
	 remote protocol does not allow this.  */
      if ((*the_low_target.breakpoint_at) (stop_pc))
	{
	  if (debug_threads)
	    fprintf (stderr, "%ld: hit user breakpoint.\n", 
		     GET_LWP (thread_to_gdb_id (current_inferior)));

	  process->pending_is_breakpoint = 1;
	  process->pending_stop_pc = stop_pc;
	  return wstat;
	}

      if (debug_threads)
	fprintf (stderr, "%ld: none of the above, hmm....\n", 
		 GET_LWP (thread_to_gdb_id (current_inferior)));
      return wstat;
    }

  /* NOTREACHED */
  return 0;
}

unsigned char
fbsd_wait (char *status)
{
  int w;

  enable_async_io ();
  unblock_async_io ();
  w = fbsd_wait_for_event ();
  disable_async_io ();

  /* If we are waiting for a particular child, and it exited,
     fbsd_wait_for_event will return its exit status.  Similarly if
     the last child exited.  If this is not the last child, however,
     do not report it as exited until there is a 'thread exited' response
     available in the remote protocol.  Instead, just wait for another 
     event. This should be safe, because if the thread crashed we will 
     already have reported the termination signal to GDB; that should 
     stop any in-progress stepping operations, etc. */
  if (WIFEXITED (w))
    {
      fprintf (stderr, "\nChild exited with retcode = %x \n", WEXITSTATUS (w));
      *status = 'W';
      clear_inferiors ();
      return ((unsigned char) WEXITSTATUS (w));
    }
  else if (!WIFSTOPPED (w))
    {
      fprintf (stderr, "\nChild terminated with signal = %x \n", WTERMSIG (w));
      *status = 'X';
      clear_inferiors ();
      return ((unsigned char) WTERMSIG (w));
    }

  *status = 'T';
  return ((unsigned char) WSTOPSIG (w));
}

static int
regsets_fetch_inferior_registers (void)
{
  struct regset_info *regset;
  regset = target_regsets;

  if (debug_threads)
    fprintf (stderr, "%ld: fetch registers.\n", 
	     GET_LWP (thread_to_gdb_id (current_inferior)));

  while (regset->size >= 0)
    {
      void *buf;
      int res;

      if (regset->size == 0)
	{
	  regset ++;
	  continue;
	}

      buf = malloc (regset->size);
      res = ptrace (regset->get_request, 
		    GET_LWP (thread_to_gdb_id (current_inferior)),
		    (PTRACE_ARG3_TYPE) buf, 0);
      if (res < 0)
	{
	  char s[256];
	  sprintf (s, "ptrace(regsets_fetch_inferior_registers) PID=%d",
		   GET_LWP (thread_to_gdb_id (current_inferior)));
	  perror (s);
	}

      if (debug_threads && regset->type == GENERAL_REGS)
	print_grs ((CORE_ADDR*) buf, 
		   GET_LWP (thread_to_gdb_id (current_inferior)));

      regset->store_function (buf);
      regset ++;
      free (buf);
    }
  return 0;
}

static int
regsets_store_inferior_registers (void)
{
  struct regset_info *regset;
  regset = target_regsets;

  if (debug_threads)
    fprintf (stderr, "%ld: store registers.\n", 
	     GET_LWP (thread_to_gdb_id (current_inferior)));

  while (regset->size >= 0)
    {
      void *buf;
      int res;

      if (regset->size == 0)
	{
	  regset ++;
	  continue;
	}

      buf = malloc (regset->size);

      /* First fill the buffer with the current register set contents,
	 in case there are any items in the kernel's regset that are
	 not in gdbserver's regcache.  */
      res = ptrace (regset->get_request, 
		    GET_LWP (thread_to_gdb_id (current_inferior)),
		    (PTRACE_ARG3_TYPE) buf, 0);

      if (res == 0)
	{
	  if (debug_threads && regset->type == GENERAL_REGS)
	    print_grs ((CORE_ADDR*) buf, 
		       GET_LWP (thread_to_gdb_id (current_inferior)));

	  /* Then overlay our cached registers on that.  */
	  regset->fill_function (buf);

	  if (debug_threads && regset->type == GENERAL_REGS)
	    print_grs ((CORE_ADDR*) buf, 
		       GET_LWP (thread_to_gdb_id (current_inferior)));

	  /* Only now do we write the register set.  */
	  res = ptrace (regset->set_request, 
			GET_LWP (thread_to_gdb_id (current_inferior)),
			(PTRACE_ARG3_TYPE) buf, 0);
	}

      if (res < 0)
	{
	  perror ("Warning: ptrace(regsets_store_inferior_registers)");
	}
      regset ++;
      free (buf);
    }
  return 0;
}

void
fbsd_fetch_registers (int regno)
{
  regsets_fetch_inferior_registers ();
}

void
fbsd_store_registers (int regno)
{
  regsets_store_inferior_registers ();
}


/* Copy LEN bytes from inferior's memory starting at MEMADDR
   to debugger memory starting at MYADDR.  */

static int
fbsd_read_memory (CORE_ADDR memaddr, unsigned char *myaddr, int len)
{
  register int i;
  /* Round starting address down to longword boundary.  */
  register CORE_ADDR addr = memaddr & -(CORE_ADDR) sizeof (PTRACE_XFER_TYPE);
  /* Round ending address up; get number of longwords that makes.  */
  register int count
    = (((memaddr + len) - addr) + sizeof (PTRACE_XFER_TYPE) - 1)
      / sizeof (PTRACE_XFER_TYPE);
  /* Allocate buffer of that many longwords.  */
  register PTRACE_XFER_TYPE *buffer
    = (PTRACE_XFER_TYPE *) alloca (count * sizeof (PTRACE_XFER_TYPE));

  /* Read all the longwords */
  for (i = 0; i < count; i++, addr += sizeof (PTRACE_XFER_TYPE))
    {
      errno = 0;
      buffer[i] = ptrace (PT_READ_D, 
			  ptid_get_pid (thread_to_gdb_id (current_inferior)), 
			  (PTRACE_ARG3_TYPE) (intptr_t) addr, 0);
      if (errno)
	return errno;
    }

  /* Copy appropriate bytes out of the buffer.  */
  memcpy (myaddr, 
	  (char *) buffer + (memaddr & (sizeof (PTRACE_XFER_TYPE) - 1)), len);

  return 0;
}

/* Copy LEN bytes of data from debugger memory at MYADDR
   to inferior's memory at MEMADDR.
   On failure (cannot write the inferior)
   returns the value of errno.  */

static int
fbsd_write_memory (CORE_ADDR memaddr, const unsigned char *myaddr, int len)
{
  extern int errno;
  register int i;

  /* Round starting address down to longword boundary.  */
  register CORE_ADDR addr = memaddr & - (CORE_ADDR) sizeof (PTRACE_XFER_TYPE);

  /* Round ending address up; get number of longwords that makes.  */
  register int count = 
    (((memaddr + len) - addr) + sizeof (PTRACE_XFER_TYPE) - 1) / 
    sizeof (PTRACE_XFER_TYPE);

  /* Allocate buffer of that many longwords.  */
  register PTRACE_XFER_TYPE *buffer = 
    (PTRACE_XFER_TYPE *) alloca (count * sizeof (PTRACE_XFER_TYPE));

  if (len && debug_threads)
    {
      fprintf (stderr, "Writing %02x to %08lx\n", 
	       (unsigned) myaddr[0], (long) memaddr);
    }

  /* Fill start and end extra bytes of buffer with existing memory data.  */
  buffer[0] = ptrace (PT_READ_D, 
		      ptid_get_pid (thread_to_gdb_id (current_inferior)),
		      (PTRACE_ARG3_TYPE) (intptr_t) addr, 0);

  if (count > 1)
    {
      buffer[count - 1]
	= ptrace (PT_READ_D, 
		  ptid_get_pid(thread_to_gdb_id(current_inferior)),
		  (PTRACE_ARG3_TYPE) (intptr_t) (addr + (count - 1)
				      * sizeof (PTRACE_XFER_TYPE)),
		  0);
    }

  /* Copy data to be written over corresponding part of buffer */
  memcpy ((char *) buffer + (memaddr & (sizeof (PTRACE_XFER_TYPE) - 1)), 
	  myaddr, len);

  /* Write the entire buffer.  */
  for (i = 0; i < count; i++, addr += sizeof (PTRACE_XFER_TYPE))
    {
      errno = 0;
      ptrace (PT_WRITE_D, 
	      ptid_get_pid (thread_to_gdb_id (current_inferior)), 
	      (PTRACE_ARG3_TYPE) (intptr_t) addr, buffer[i]);
      if (errno)
	return errno;
    }

  return 0;
}

static void
fbsd_look_up_symbols (void)
{
#ifdef USE_THREAD_DB
  if (using_threads)
    return;

  using_threads = 
    thread_db_init (ptid_get_pid (thread_to_gdb_id (current_inferior)));
#endif
}

static void
fbsd_send_signal (int signum)
{
  extern int signal_pid;

  if (!ptid_equal (cont_thread, null_ptid) && 
      !ptid_equal (cont_thread, minus_one_ptid))
    {
      kill (ptid_get_lwp (thread_to_gdb_id (current_inferior)), signum);
    }
  else
    {
      kill (signal_pid, signum);
    }
}

/* Copy LEN bytes from inferior's auxiliary vector starting at OFFSET
   to debugger memory starting at MYADDR.  */
static int
fbsd_read_auxv (CORE_ADDR offset, unsigned char *myaddr, unsigned int len)
{
  char filename[PATH_MAX];
  int fd, n;

  snprintf (filename, sizeof filename, "/proc/%d/auxv", 
	    ptid_get_pid (thread_to_gdb_id (current_inferior)));

  fd = open (filename, O_RDONLY);
  if (fd < 0)
    return -1;

  if (offset != (CORE_ADDR) 0
      && lseek (fd, (off_t) offset, SEEK_SET) != (off_t) offset)
    n = -1;
  else
    n = read (fd, myaddr, len);

  close (fd);

  return n;
}

/* These watchpoint related wrapper functions simply pass on the function call
   if the target has registered a corresponding function.  */
static int
fbsd_insert_watchpoint (char type, CORE_ADDR addr, int len)
{
  if (the_low_target.insert_watchpoint != NULL)
    return the_low_target.insert_watchpoint (type, addr, len);
  else
    /* Unsupported (see target.h).  */
    return 1;
}

static int
fbsd_remove_watchpoint (char type, CORE_ADDR addr, int len)
{
  if (the_low_target.remove_watchpoint != NULL)
    return the_low_target.remove_watchpoint (type, addr, len);
  else
    /* Unsupported (see target.h).  */
    return 1;
}

static int
fbsd_stopped_by_watchpoint (void)
{
  if (the_low_target.stopped_by_watchpoint != NULL)
    return the_low_target.stopped_by_watchpoint ();
  else
    return 0;
}

static CORE_ADDR
fbsd_stopped_data_address (void)
{
  if (the_low_target.stopped_data_address != NULL)
    return the_low_target.stopped_data_address ();
  else
    return 0;
}


static struct target_ops fbsd_target_ops = {
  fbsd_create_inferior,
  fbsd_attach,
  fbsd_kill,
  fbsd_detach,
  fbsd_thread_alive,
  fbsd_resume,
  fbsd_wait,
  fbsd_fetch_registers,
  fbsd_store_registers,
  fbsd_read_memory,
  fbsd_write_memory,
  fbsd_look_up_symbols,
  fbsd_send_signal,
  fbsd_read_auxv,
  fbsd_insert_watchpoint,
  fbsd_remove_watchpoint,
  fbsd_stopped_by_watchpoint,
  fbsd_stopped_data_address,
  NULL,
#ifdef USE_THREAD_DB
  thread_db_get_tls_address,
#else
  NULL,
#endif
};

void
initialize_low (void)
{
  using_threads = 0;
  set_target_ops (&fbsd_target_ops);
  set_breakpoint_data (the_low_target.breakpoint,
		       the_low_target.breakpoint_len);
  init_registers ();
}
