/*
 * Copyright (c) 2004 Marcel Moolenaar
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD: src/gnu/usr.bin/gdb/kgdb/kthr.c,v 1.2.2.7 2008/02/25 22:20:13 jhb Exp $");

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <err.h>
#include <inttypes.h>
#include <kvm.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <defs.h>
#include <frame-unwind.h>

#include "kgdb.h"

static struct kthr *first;
struct kthr *curkthr;
CORE_ADDR dumppcb;
static CORE_ADDR stoppcbs;

/* cpumask_t was deprecated in FreeBSD 10, but our kgdb needs to support
   both legacy Junos and FreeBSD 10 based Junos. Replacing the use of
   cpumask_t with an explicitly sized uint. */
#if defined(__powerpc64__)
static uint64_t stopped_cpus;
#else
static uint32_t stopped_cpus;
#endif

static size_t p_pid_off, pthr_tqh_first_off, plist_le_next_off, td_state_off,
  td_oncpu_off, td_kstack_off, td_tid_off, td_pcb_off, td_plist_tqe_next_off, 
  sizeof_proc, sizeof_thread, sizeof_ptr, sizeof_int, sizeof_long, sizeof_pcb,
  sizeof_cpuset;
static unsigned int dumptid;
static int td_state;
static char *cpuset_ptr;

static void
kgdb_thr_add_procs(CORE_ADDR paddr)
{
	struct kthr *kt;
	unsigned int pid;
	CORE_ADDR taddr;
	char *proc_ptr, *thread_ptr;
	int thread_on_cpu, cpuword_idx;
	long cpuword, cpuset_mask;

	proc_ptr = alloca(sizeof_proc);
	thread_ptr = alloca(sizeof_thread);

	while (paddr != 0) {
		if (kvm_read(kvm, paddr, proc_ptr, sizeof_proc) != 
		    sizeof_proc) {
			warnx("kvm_read: %s", kvm_geterr(kvm));
			break;
		}
		pid = extract_unsigned_integer(
			proc_ptr + p_pid_off, sizeof_int);

		taddr = extract_signed_integer(
			proc_ptr + pthr_tqh_first_off, sizeof_ptr);
		while (taddr != 0) {
			if (kvm_read(kvm, taddr, thread_ptr, sizeof_thread) !=
			    sizeof_thread) {
				warnx("kvm_read: %s", kvm_geterr(kvm));
				break;
			}

			kt = malloc(sizeof(*kt));
			kt->next = first;
			first = kt;

			td_state = extract_signed_integer(
				thread_ptr + td_state_off, sizeof_int);
			kt->cpu = extract_signed_integer(
				thread_ptr + td_oncpu_off, 1);
			kt->kaddr = taddr;
			kt->kstack = extract_signed_integer(
				thread_ptr + td_kstack_off, sizeof_ptr);
			kt->paddr = paddr;
			kt->pid = pid;
			kt->tid = extract_signed_integer(
				thread_ptr + td_tid_off, sizeof_int);

			if (cpuset_ptr) {
				cpuword_idx = 
					(sizeof_cpuset == sizeof_long) ? 
					0 : kt->cpu/(sizeof_long * 8);
				cpuword = *((long*) (cpuset_ptr + 
						     cpuword_idx * sizeof_long));
				cpuset_mask = (long) 1 << 
					((sizeof_cpuset == sizeof_long) ? 
					 kt->cpu : kt->cpu/(sizeof_long * 8));
				thread_on_cpu = cpuword & cpuset_mask;
			}
			else {
				thread_on_cpu = ((1 << kt->cpu) & stopped_cpus);
			}

			if (kt->tid == dumptid)
			        kt->pcb = dumppcb;
			else if (td_state == TDS_RUNNING && thread_on_cpu &&
				 stoppcbs != 0)
				kt->pcb = stoppcbs + sizeof_pcb * kt->cpu;
			else
				kt->pcb = extract_signed_integer(
					thread_ptr + td_pcb_off, sizeof_ptr);

			taddr = extract_signed_integer(
				thread_ptr + td_plist_tqe_next_off, sizeof_ptr);
		}

		paddr = extract_signed_integer(
			proc_ptr + plist_le_next_off, sizeof_ptr);
	}
}

struct kthr *
kgdb_thr_init(void)
{
	struct kthr *kt;
	CORE_ADDR paddr, addr;

	while (first != NULL) {
		kt = first;
		first = kt->next;
		free(kt);
	}

	dumptid = kgdb_parse("dumptid");
	dumppcb = kgdb_parse("&dumppcb");
	stoppcbs = kgdb_parse("&stoppcbs");

	p_pid_off = kgdb_parse("&((struct proc *)0)->p_pid");
	pthr_tqh_first_off = 
		kgdb_parse("&((struct proc *)0)->p_threads.tqh_first");
	plist_le_next_off = kgdb_parse("&((struct proc *)0)->p_list.le_next");

	td_state_off = kgdb_parse("&((struct thread *)0)->td_state");
	td_oncpu_off = kgdb_parse("&((struct thread *)0)->td_oncpu");
	td_kstack_off = kgdb_parse("&((struct thread *)0)->td_kstack");
	td_tid_off = kgdb_parse("&((struct thread *)0)->td_tid");
	td_pcb_off = kgdb_parse("&((struct thread *)0)->td_pcb");
	td_plist_tqe_next_off = 
		kgdb_parse("&((struct thread *)0)->td_plist.tqe_next");

	sizeof_proc = kgdb_parse("sizeof (struct proc)");
	sizeof_thread = kgdb_parse("sizeof (struct thread)");
	sizeof_ptr = kgdb_parse("sizeof (void*)");
	sizeof_int = kgdb_parse("sizeof (int)");
	sizeof_long = kgdb_parse("sizeof (long)");
	sizeof_pcb = kgdb_parse("sizeof(struct pcb)");
	sizeof_cpuset = kgdb_parse("sizeof(cpuset_t)");

	/* Read stopped cpus. The type of stopped_cpus is different on 
	   FreeBSD 9. Need to handle both. */
	volatile struct gdb_exception except;
	TRY_CATCH (except, RETURN_MASK_ERROR) {
		stopped_cpus = kgdb_parse("stopped_cpus");
	}
	/* If parse fails it's because stopped_cpus is an aggregate type. */
	cpuset_ptr = NULL;
	if (except.reason < 0) {
		addr = kgdb_parse("&stopped_cpus");
		/* Read the cpuset array. */
		if (addr) {
			cpuset_ptr = malloc(sizeof_cpuset);
			if (kvm_read(kvm, addr, cpuset_ptr, sizeof_cpuset) != 
			    sizeof_cpuset) {
				warnx("kvm_read: %s", kvm_geterr(kvm));
			}
		}
	}

	paddr = kgdb_parse("allproc.lh_first");
	kgdb_thr_add_procs(paddr);

	paddr = kgdb_parse("zombproc.lh_first");
	if (paddr != 0) {
		kgdb_thr_add_procs(paddr);
	}

	curkthr = kgdb_thr_lookup_tid(dumptid);
	if (curkthr == NULL)
		curkthr = first;

	if (cpuset_ptr == NULL)
		free(cpuset_ptr);

	return (first);
}

struct kthr *
kgdb_thr_lookup_tid(int tid)
{
	struct kthr *kt;

	kt = first;
	while (kt != NULL && kt->tid != tid)
		kt = kt->next;
	return (kt);
}

struct kthr *
kgdb_thr_lookup_taddr(CORE_ADDR taddr)
{
	struct kthr *kt;

	kt = first;
	while (kt != NULL && kt->kaddr != taddr)
		kt = kt->next;
	return (kt);
}

struct kthr *
kgdb_thr_lookup_pid(int pid)
{
	struct kthr *kt;

	kt = first;
	while (kt != NULL && kt->pid != pid)
		kt = kt->next;
	return (kt);
}

struct kthr *
kgdb_thr_lookup_paddr(CORE_ADDR paddr)
{
	struct kthr *kt;

	kt = first;
	while (kt != NULL && kt->paddr != paddr)
		kt = kt->next;
	return (kt);
}

struct kthr *
kgdb_thr_next(struct kthr *kt)
{
	return (kt->next);
}

struct kthr *
kgdb_thr_select(struct kthr *kt)
{
	struct kthr *pcur;

	pcur = curkthr;
	curkthr = kt;
	return (pcur);
}

char *
kgdb_thr_extra_thread_info(int tid)
{
	static char comm[MAXCOMLEN + 1];
	static char buf[64];
	struct kthr *kt;
	CORE_ADDR off_p_comm;

	kt = kgdb_thr_lookup_tid(tid);
	if (kt == NULL)
		return (NULL);
	snprintf(buf, sizeof(buf), "PID=%d", kt->pid);
	off_p_comm = kgdb_parse("&((struct proc *)0)->p_comm[0]");
	if (kvm_read(kvm, kt->paddr + off_p_comm, &comm, sizeof(comm)) !=
	    sizeof(comm))
		return (buf);

	strlcat(buf, ": ", sizeof(buf));
	strlcat(buf, comm, sizeof(buf));
	return (buf);
}
