/*
 * Copyright (c) 2004 Marcel Moolenaar
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD: src/gnu/usr.bin/gdb/kgdb/trgt_amd64.c,v 1.2.2.6 2008/05/09 18:59:50 jhb Exp $");

#include <sys/types.h>
#include <err.h>
#include <kvm.h>
#include <string.h>

#include <defs.h>
#include <target.h>
#include <gdbthread.h>
#include <inferior.h>
#include <regcache.h>
#include <frame-unwind.h>
#include <amd64-tdep.h>

#include "kgdb.h"

static size_t pcb_rbx_off, pcb_rbp_off, pcb_rsp_off, pcb_r12_off, pcb_r13_off, 
  pcb_r14_off, pcb_r15_off, pcb_rip_off, pcb_size;

void
kgdb_trgt_fetch_registers(int regno __unused)
{
	struct kthr *kt;
	char *pcb;

	kt = kgdb_thr_lookup_tid(ptid_get_pid(inferior_ptid));
	if (kt == NULL || pcb_size == 0)
		return;
	pcb = alloca(pcb_size);
	if (pcb == NULL)
		return;
	if (kvm_read(kvm, kt->pcb, pcb, pcb_size) != pcb_size) {
		warnx("kvm_read: %s", kvm_geterr(kvm));
		memset(pcb, 0, pcb_size);
	}

	regcache_raw_supply(current_regcache, AMD64_RBX_REGNUM, 
			    pcb + pcb_rbx_off);
	regcache_raw_supply(current_regcache, AMD64_RBP_REGNUM, 
			    pcb + pcb_rbp_off);
	regcache_raw_supply(current_regcache, AMD64_RSP_REGNUM, 
			    pcb + pcb_rsp_off);
	regcache_raw_supply(current_regcache, AMD64_R8_REGNUM + 4, 
			    pcb + pcb_r12_off);
	regcache_raw_supply(current_regcache, AMD64_R8_REGNUM + 5, 
			    pcb + pcb_r13_off);
	regcache_raw_supply(current_regcache, AMD64_R8_REGNUM + 6, 
			    pcb + pcb_r14_off);
	regcache_raw_supply(current_regcache, AMD64_R15_REGNUM, 
			    pcb + pcb_r15_off);
	regcache_raw_supply(current_regcache, AMD64_RIP_REGNUM, 
			    pcb + pcb_rip_off);
}

void
kgdb_trgt_store_registers(int regno __unused)
{
	fprintf_unfiltered(gdb_stderr, "XXX: %s\n", __func__);
}

struct kgdb_frame_cache {
	CORE_ADDR	pc;
	CORE_ADDR	sp;
};
static int kgdb_trgt_frame_offset[20];

void
kgdb_trgt_init_type_info(void)
{
	pcb_size = kgdb_parse("sizeof(struct pcb)");
	pcb_rbx_off = kgdb_parse("&((struct pcb *)0)->pcb_rbx");
	pcb_rbp_off = kgdb_parse("&((struct pcb *)0)->pcb_rbp");
	pcb_rsp_off = kgdb_parse("&((struct pcb *)0)->pcb_rsp");
	pcb_r12_off = kgdb_parse("&((struct pcb *)0)->pcb_r12");
	pcb_r13_off = kgdb_parse("&((struct pcb *)0)->pcb_r13");
	pcb_r14_off = kgdb_parse("&((struct pcb *)0)->pcb_r14");
	pcb_r15_off = kgdb_parse("&((struct pcb *)0)->pcb_r15");
	pcb_rip_off = kgdb_parse("&((struct pcb *)0)->pcb_rip");

	kgdb_trgt_frame_offset[0] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_rax");
	kgdb_trgt_frame_offset[1] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_rbx");
	kgdb_trgt_frame_offset[2] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_rcx");
	kgdb_trgt_frame_offset[3] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_rdx");
	kgdb_trgt_frame_offset[4] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_rsi");
	kgdb_trgt_frame_offset[5] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_rdi");
	kgdb_trgt_frame_offset[6] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_rbp");
	kgdb_trgt_frame_offset[7] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_rsp");
	kgdb_trgt_frame_offset[8] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_r8");
	kgdb_trgt_frame_offset[9] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_r9");
	kgdb_trgt_frame_offset[10] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_r10");
	kgdb_trgt_frame_offset[11] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_r11");
	kgdb_trgt_frame_offset[12] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_r12");
	kgdb_trgt_frame_offset[13] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_r13");
	kgdb_trgt_frame_offset[14] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_r14");
	kgdb_trgt_frame_offset[15] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_r15");
	kgdb_trgt_frame_offset[16] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_rip");
	kgdb_trgt_frame_offset[17] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_rflags");
	kgdb_trgt_frame_offset[18] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_cs");
	kgdb_trgt_frame_offset[19] = 
	  kgdb_parse("&((struct trapframe*)0)->tf_ss");
}

static struct kgdb_frame_cache *
kgdb_trgt_frame_cache(struct frame_info *next_frame, void **this_cache)
{
	char buf[MAX_REGISTER_SIZE];
	struct kgdb_frame_cache *cache;

	cache = *this_cache;
	if (cache == NULL) {
		cache = FRAME_OBSTACK_ZALLOC(struct kgdb_frame_cache);
		*this_cache = cache;
		cache->pc = frame_func_unwind(next_frame);
		frame_unwind_register(next_frame, SP_REGNUM, buf);
		cache->sp = extract_unsigned_integer(buf,
		    register_size(current_gdbarch, SP_REGNUM));
	}
	return (cache);
}

static void
kgdb_trgt_trapframe_this_id(struct frame_info *next_frame, void **this_cache,
    struct frame_id *this_id)
{
	struct kgdb_frame_cache *cache;

	cache = kgdb_trgt_frame_cache(next_frame, this_cache);
	*this_id = frame_id_build(cache->sp, cache->pc);
}

static void
kgdb_trgt_trapframe_prev_register(struct frame_info *next_frame,
    void **this_cache, int regnum, int *optimizedp, enum lval_type *lvalp,
    CORE_ADDR *addrp, int *realnump, void *valuep)
{
	char dummy_valuep[MAX_REGISTER_SIZE];
	struct kgdb_frame_cache *cache;
	int ofs, regsz;

	regsz = register_size(current_gdbarch, regnum);

	if (valuep == NULL)
		valuep = dummy_valuep;
	memset(valuep, 0, regsz);
	*optimizedp = 0;
	*addrp = 0;
	*lvalp = not_lval;
	*realnump = -1;

	ofs = (regnum >= AMD64_RAX_REGNUM && regnum <= AMD64_EFLAGS_REGNUM + 2)
	    ? kgdb_trgt_frame_offset[regnum] : -1;
	if (ofs == -1)
		return;

	cache = kgdb_trgt_frame_cache(next_frame, this_cache);
	*addrp = cache->sp + ofs;
	*lvalp = lval_memory;
	target_read_memory(*addrp, valuep, regsz);
}

/* FIXME: is NORMAL_FRAME acceptable for this? */
/* FIXME: are the last three initialized correctly? */
static const struct frame_unwind kgdb_trgt_trapframe_unwind = {
    NORMAL_FRAME,
    &kgdb_trgt_trapframe_this_id,
    &kgdb_trgt_trapframe_prev_register,
    NULL, /*  const struct frame_data *unwind_data; */
    &kgdb_trgt_trapframe_sniffer, /*  frame_sniffer_ftype *sniffer;         */
    NULL  /*  frame_prev_pc_ftype *prev_pc;         */
};

const struct frame_unwind *
kgdb_trgt_trapframe_sniffer(struct frame_info *next_frame)
{
	char *pname;
	CORE_ADDR pc;

	pc = frame_pc_unwind(next_frame);
	pname = NULL;
	find_pc_partial_function(pc, &pname, NULL, NULL);
	if (pname == NULL)
		return (NULL);
	if (strcmp(pname, "calltrap") == 0 ||
	    strcmp(pname, "nmi_calltrap") == 0 ||
	    (pname[0] == 'X' && pname[1] != '_'))
		return (&kgdb_trgt_trapframe_unwind);
	/* printf("%s: %lx =%s\n", __func__, pc, pname); */
	return (NULL);
}
